from enum import Enum


class ColumnSalience(Enum):
    """An enum that holds the salience levels options that can be used when
    specifying data columns. Salience is used to indicate whether it would be best to
    store the data in the data store or whether it can be just stored in
    the local file-system and discarded after it has been used. This choice
    is ultimately specified by the user by defining a salience threshold for
    a store.

    The salience is also used when providing information on what sinks
    are available to avoid cluttering help menus
    """

    primary = (
        100,
        "Primary input data, typically reconstructed by the instrument that "
        "collects them",
    )
    raw = (
        90,
        "Raw data from the scanner that haven't been reconstructed and are "
        "only typically used in advanced analyses",
    )
    publication = (
        80,
        "Results that would typically be used as main outputs in publications",
    )
    supplementary = (
        60,
        "Derivatives that would typically only be provided in "
        "supplementary material",
    )
    qa = (
        40,
        "Derivatives that would typically be only kept for quality "
        "assurance of analysis workflows",
    )
    debug = (
        20,
        "Derivatives that would typically only need to be checked "
        "when debugging analysis workflows",
    )
    temp = (
        0,
        "Data only temporarily stored to pass between pipelines, e.g. that "
        "operate on different row frequencies",
    )

    def __init__(self, level, desc):
        self.level = level
        self.desc = desc

    def __lt__(self, other):
        return self.level < other.level

    def __le__(self, other):
        return self.level <= other.level

    def __str__(self):
        return self.name

    @classmethod
    def default(self):
        return self.supplementary


class ParameterSalience(Enum):
    """An enum that holds the salience levels options that can be used when
    specifying class parameters. Salience is used to indicate whether the
    parameter should show up by default when listing the available parameters
    of an Analysis class in a menu.
    """

    def __str__(self):
        return self.name

    debug = (0, "typically only needed to be altered for debugging")
    recommended = (20, "recommended to keep defaults")
    dependent = (
        40,
        "best value can be dependent on the context of the analysis, but the default "
        "should work for most cases",
    )
    check = (60, "default value should be checked for validity for particular use case")
    arbitrary = (80, "a default is provided, but it is not clear which value is best")
    required = (100, "No sensible default value, should be provided")

    @classmethod
    def default(self):
        return self.recommended


class CheckSalience(Enum):
    """An enum that holds the potential values for signifying how critical a check is to
    run.
    """

    def __str__(self):
        return self.name

    debug = (0, "typically only used to debug alterations to the pipeline")
    potential = (20, "check can be run but not typically necessary")
    prudent = (
        40,
        "it is prudent to run the check the results but you can skip if required",
    )
    recommended = (
        60,
        "recommended to run the check as pipeline fails 1~2% of the time",
    )
    strongly_recommended = (
        80,
        "strongly recommended to run the check as pipeline fails 5~10% of the time",
    )
    required = (100, "Pipeline will often fail, checking the results is required")

    @classmethod
    def default(self):
        return self.recommended


class CheckStatus(Enum):
    """An enum that holds the potential values that signify how likely a pipeline has "
    "failed"""

    def __str__(self):
        return self.name

    failed = (0, "the pipeline has failed")
    probable_fail = (25, "probable that the pipeline has failed")
    unclear = (
        50,
        "cannot ascertain whether the pipeline has failed or not",
    )
    probable_pass = (75, "probable that the pipeline has run successfully")
    passed = (100, "the pipeline has run successfully")

    @classmethod
    def default(self):
        return self.unclear



from arcana.core.cli.ext import ext
from arcana.core.utils.misc import show_cli_trace


def test_cli_ext(cli_runner):

    result = cli_runner(ext, ["--help"])

    assert result.exit_code == 0, show_cli_trace(result)



from __future__ import annotations
import typing as ty
import inspect
import itertools
from copy import copy
from operator import attrgetter
import attrs
from arcana.core.data.column import DataColumn
from arcana.core.data.set import Dataset
from arcana.core.utils.misc import (
    PIPELINE_ANNOTATIONS,
    SWICTH_ANNOTATIONS,
    CHECK_ANNOTATIONS,
)
from arcana.core.exceptions import ArcanaDesignError
from .spec import (
    BaseAttr,
    ColumnSpec,
    Parameter,
    BaseMethod,
    SubanalysisSpec,
    PipelineConstructor,
    Switch,
    Check,
    AnalysisSpec,
    _Inherited,
    _MappedFrom,
)


RESERVED_NAMES = ("dataset", "menu", "stack")
RESERVED_NAMES += tuple("_" + n for n in RESERVED_NAMES)


@attrs.define(kw_only=True, slots=False)
class Analysis:

    name: str


def make(klass: type, space: type) -> type:
    """
    Construct an analysis class and validate all the components fit together

    Parameters
    ----------
    klass : type
        the class that is decorated with the @analysis decorator
    space : type
        a subclass of the DataSpace enum, which defines the types of
        datasets the analysis can be applied to

    Returns
    -------
    type
        the analysis class
    """

    # Initialise lists to hold all the different components of an analysis
    column_specs = []
    pipeline_builders = []
    parameters = []
    switches = []
    checks = []
    subanalysis_specs = []

    # Set name and datatype of attributes and Resolve 'Inherited' and 'mapped_from'
    # attributes and create list of columns, parameters and subanalyses
    for name, attr in list(klass.__dict__.items()):
        if name in RESERVED_NAMES:
            raise ArcanaDesignError(
                f"Cannot use reserved name '{name}' for attribute in "
                f"'{klass.__name__}' analysis class"
            )
        # Get attribute type from type annotations
        dtype = klass.__annotations__.get(name)
        if ty.get_origin(dtype) is DataColumn:
            dtype = ty.get_args(dtype)[0]
        # Resolve inherited and mapped attributes
        if isinstance(attr, (_Inherited, _MappedFrom)):
            resolved, resolved_dtype = attr.resolve(name, klass)
            if dtype is None:
                # Copy type annotation across to new class if it isn't present
                dtype = resolved_dtype
            elif dtype is not resolved_dtype and not issubclass(dtype, resolved_dtype):
                raise ArcanaDesignError(
                    f"Cannot change datatype of {name} from {resolved_dtype} to "
                    f"{dtype} as it is not a sub-class"
                )
            attr.resolved_to = resolved
            attr = resolved

        if isinstance(attr, BaseAttr):
            # Save annotated type of column in metadata and convert to Column
            if dtype is None:
                raise ArcanaDesignError(
                    f"Type annotation must be provided for '{name}' {type(attr).__name__}"
                )
            # Set the name and type of the attributes from the class dict and annotaitons
            # respectively. Need to use object setattr to avoid frozen status
            object.__setattr__(attr, "name", name)
            object.__setattr__(attr, "type", dtype)
            if isinstance(attr, ColumnSpec):
                if attr.row_frequency is None:
                    object.__setattr__(
                        attr, "row_frequency", max(space)
                    )  # "Leaf" frequency of the data tree
                column_specs.append(attr)
            elif isinstance(attr, Parameter):
                parameters.append(attr)
            elif isinstance(attr, SubanalysisSpec):
                subanalysis_specs.append(attr)

    # Resolve the mappings from through the subanalysis_specs in a separate loop so the
    # column names can be resolved
    for spec in subanalysis_specs:
        resolved_mappings = []
        for (from_, to) in spec.mappings:
            resolved_mappings.append((from_, to.name))
        # Add in implicit mappings, where a column from the subanalysis has been
        # mapped into the global namespace of the analysis class
        for col_or_param in itertools.chain(column_specs, parameters):
            if col_or_param.mapped_from and col_or_param.mapped_from[0] == spec.name:
                resolved_mappings.append(
                    (col_or_param.mapped_from[1], col_or_param.name)
                )
        object.__setattr__(spec, "mappings", tuple(sorted(resolved_mappings)))

    # Attributes that need to be converted into attrs.fields before the class
    # is attrisfied
    to_convert_to_attrs = column_specs + parameters + subanalysis_specs

    # Loop through all attributes to pick up decorated methods for pipelines, checks
    # and switches
    for attr in klass.__dict__.values():
        try:
            attr_anots = attr.__annotations__
        except AttributeError:
            continue

        if PIPELINE_ANNOTATIONS in attr_anots:
            anots = attr_anots[PIPELINE_ANNOTATIONS]
            outputs = tuple(o.name for o in anots["outputs"])
            input_columns, used_parameters = _get_args_automagically(
                column_specs=column_specs, parameters=parameters, method=attr
            )
            unresolved_condition = anots["condition"]
            if unresolved_condition is not None:
                try:
                    condition = unresolved_condition.name
                except AttributeError:
                    condition = unresolved_condition.resolve(
                        klass,
                        column_specs=column_specs,
                        parameters=parameters,
                    )
            else:
                condition = None
            pipeline_builders.append(
                PipelineConstructor(
                    name=attr.__name__,
                    desc=attr.__doc__,
                    inputs=input_columns,
                    outputs=outputs,
                    parameters=used_parameters,
                    condition=condition,
                    switch=anots["switch"],
                    defined_in=(),
                    method=attr,
                )
            )
        elif SWICTH_ANNOTATIONS in attr_anots:
            input_columns, used_parameters = _get_args_automagically(
                column_specs=column_specs, parameters=parameters, method=attr
            )
            switches.append(
                Switch(
                    name=attr.__name__,
                    desc=__doc__,
                    inputs=input_columns,
                    parameters=used_parameters,
                    defined_in=(),
                    method=attr,
                )
            )
        elif CHECK_ANNOTATIONS in attr_anots:
            anots = attr_anots[CHECK_ANNOTATIONS]
            column_name = anots["column"].name
            input_columns, used_parameters = _get_args_automagically(
                column_specs=column_specs, parameters=parameters, method=attr
            )
            checks.append(
                Check(
                    name=attr.__name__,
                    column=column_name,
                    desc=attr.__doc__,
                    inputs=input_columns,
                    parameters=used_parameters,
                    salience=anots["salience"],
                    defined_in=(),
                    method=attr,
                )
            )

    to_set_defined_in = to_convert_to_attrs + pipeline_builders + switches + checks

    # Combine with specs from base classes
    for base in klass.__mro__[1:]:
        if not hasattr(base, "__spec__"):
            continue  # skip classes that aren't decorated analyses
        if base.__spec__.space is not space:  # TODO: permit "super spaces"
            raise ValueError(
                "Cannot redefine the space that an analysis operates on from "
                f"{base.__spec__.space} to {space}"
            )
        # Append column specs, parameters and subanalyses that were inherited from base
        # classes
        for lst, base_lst in (
            (column_specs, base.__spec__.column_specs),
            (parameters, base.__spec__.parameters),
            (subanalysis_specs, base.__spec__.subanalysis_specs),
        ):
            if not_inherited_explicitly := [
                x.name
                for x in lst
                if x.name in (b.name for b in base_lst) and not x.inherited
            ]:
                raise ArcanaDesignError(
                    f"{not_inherited_explicitly} attributes in {klass} implicitly override "
                    f"the corresponding attributes in {base} (i.e. without using the "
                    "inherit() function)"
                )
            lst.extend(b for b in base_lst if b.name not in (x.name for x in lst))
        # Append methods to those that that were inherited from base classes
        for lst, base_lst in (
            (pipeline_builders, base.__spec__.pipeline_builders),
            (switches, base.__spec__.switches),
            (checks, base.__spec__.checks),
        ):
            for base_method in base_lst:
                try:
                    method = next(m for m in lst if m.name == base_method.name)
                except StopIteration:
                    continue
                # Copy across defined attribute
                object.__setattr__(method, "defined_in", base_method.defined_in)
                # Check pipeline builders to see that they don't remove outputs
                if isinstance(method, PipelineConstructor):
                    if missing_outputs := [
                        o for o in base_method.outputs if o not in method.outputs
                    ]:
                        raise ArcanaDesignError(
                            f"{missing_outputs} outputs are missing from '{method.name}' "
                            "pipeline builder, which were defined by the overridden method "
                            "in {base}. Overriding methods can only add new outputs, not "
                            "remove existing ones"
                        )
            lst.extend(b for b in base_lst if b.name not in (x.name for x in lst))

    analysis_spec = AnalysisSpec(
        space=space,
        column_specs=tuple(sorted(column_specs, key=attrgetter("name"))),
        pipeline_builders=tuple(sorted(pipeline_builders, key=attrgetter("name"))),
        parameters=tuple(sorted(parameters, key=attrgetter("name"))),
        switches=tuple(sorted(switches, key=attrgetter("name"))),
        checks=tuple(sorted(checks, key=attrgetter("name"))),
        subanalysis_specs=tuple(sorted(subanalysis_specs, key=attrgetter("name"))),
    )

    # Now that we have saved the attributes in lists to be
    for attr in to_convert_to_attrs:
        attrs_field = attr.to_attrs_field()
        setattr(klass, attr.name, attrs_field)
        klass.__annotations__[attr.name] = dtype

    # Ensure that the class has it's own annotaitons dict so we can modify it without
    # messing up other classes
    klass.__annotations__ = copy(klass.__annotations__)
    klass._dataset = attrs.field(default=None, validator=_dataset_validator)
    klass.__annotations__["_dataset"] = Dataset

    # Set built-in methods
    klass.menu = MenuDescriptor()
    klass.stack = StackDescriptor()
    klass.__attrs_post_init__ = _analysis_post_init

    # Add the analysis spec to the __spec__ attribute
    klass.__spec__ = analysis_spec
    klass.__annotations__["__spec__"] = AnalysisSpec

    # Create class using attrs package, will create attributes for all columns and
    # parameters
    attrs_klass = attrs.define(auto_attribs=False)(klass)

    # Set the class that newly attribute attributes were defined in
    for attr in to_set_defined_in:
        # Register the attribute/method as being defined in this class if it has
        # been either added or created
        if isinstance(attr, BaseMethod) or len(attr.modified) == len(attr.defined_in):
            object.__setattr__(attr, "defined_in", attr.defined_in + (attrs_klass,))

    return attrs_klass


def _analysis_post_init(self):
    """Set up links to dataset columns and initialise subanalysis classes in after the
    attrs init"""
    for spec in self.__spec__.column_specs:
        column = getattr(self, spec.name)
        if isinstance(column, str):
            column = self._dataset[column]
            setattr(self, spec.name, column)
        elif column not in (None, attrs.NOTHING) and not isinstance(column, DataColumn):
            raise ValueError(
                f"Value passed to '{spec.name}' ({column}) should either be a "
                "data column or the name of a column"
            )
    for spec in self.__spec__.subanalysis_specs:
        subanalysis = getattr(self, spec.name)
        if isinstance(subanalysis, dict):
            kwargs = {m[0]: attrs.NOTHING for m in spec.mappings}
            if overwritten_mapped := list(set(kwargs) & set(subanalysis)):
                raise ValueError(
                    f"{overwritten_mapped} attributes cannot be set explicitly in '"
                    f"{spec.name}' sub-analysis as they are mapped from the enclosing "
                    "analysis"
                )
            kwargs.update(subanalysis)
            subanalysis = Subanalysis(
                spec=spec,
                analysis=spec.type(dataset=self._dataset, **kwargs),
                parent=self,
            )
            setattr(self, spec.name, subanalysis)
        elif not isinstance(subanalysis, Subanalysis):
            raise ValueError(
                f"Value passed to '{spec.name}' ({subanalysis}) should either be a "
                "subanalysis or a dict containing keyword args to initialise a "
                "subanalysis"
            )


def _get_args_automagically(column_specs, parameters, method, index_start=2):
    """Automagically determine inputs to pipeline or switched by matching
    a methods argument names with columns and parameters of the class

    Parameters
    ----------
    column_specs : list[ColumnSpec]
        the column specs to match the inputs against
    parameters : list[Parameter]
        the parameters to match the inputs against
    method : bound-method
        the method to automagically determine the inputs for
    index_start : int
        the argument index to start from (i.e. can be used to skip the workflow
        arg passed to pipelines classes)

    Returns
    -------
    ty.List[str]
        the names of the input columns to automagically provide to the method
    ty.List[str]
        the names of the parameters to automagically provide to the method
    """
    inputs = []
    used_parameters = []
    column_names = [c.name for c in column_specs]
    param_names = [p.name for p in parameters]
    signature = inspect.signature(method)
    for arg in list(signature.parameters)[
        index_start:
    ]:  # First arg is self and second is the workflow object to add to
        required_type = method.__annotations__.get(arg)
        if arg in column_names:
            column_spec = next(c for c in column_specs if c.name == arg)
            if required_type is not None and required_type is not column_spec.type:
                # Check to see whether conversion is possible
                required_type.get_converter(column_spec.type, name="dummy")
            inputs.append(arg)
        elif arg in param_names:
            used_parameters.append(arg)
        else:
            raise ArcanaDesignError(
                f"Unrecognised argument '{arg}'. If it is from a base class, "
                "make sure that it is explicitly inherited using the `Inherited` "
                "function."
            )
    return tuple(inputs), tuple(used_parameters)


def _dataset_validator(self, _, val):
    if not val:
        raise ValueError(f"A dataset must be provided when initialising {self} ")


class MenuDescriptor:
    def __get__(self, ins, typ):
        if ins is None:
            raise NotImplementedError("Class-based menu calls are not implemented")
        else:
            raise NotImplementedError("Instance-based menu calls are not implemented")


class StackDescriptor:
    def __get__(self, ins, typ):
        if ins is None:
            raise NotImplementedError("Class-based stack calls are not implemented")
        else:
            raise NotImplementedError("Instance-based stack calls are not implemented")


@attrs.define
class Subanalysis:
    """Wrapper around the actual analysis class of the subanalysis, which performs the
    mapping of attributes"""

    _spec: SubanalysisSpec
    _analysis: ty.Any  # initialised analysis class for the subanalysis
    _parent: ty.Any  # reference back to the parent analysis class

    def __repr__(self):
        return f"{type(self).__name__}(_name={self._spec.name}, type={type(self._analysis)})"

    def __getattr__(self, name: str) -> ty.Any:
        try:
            return object.__getattr__(self, name)
        except AttributeError:
            pass
        try:
            mapped_name = self._spec.mapping(name)
        except KeyError:
            value = getattr(self._analysis, name)
        else:
            value = getattr(self._parent, mapped_name)
        return value

    def __setattr__(self, name, value):
        if name in ("_spec", "_analysis", "_parent"):
            object.__setattr__(self, name, value)
        else:
            try:
                mapped_name = self._spec.mapping(name)
            except KeyError:
                pass
            else:
                raise AttributeError(
                    f"Cannot set value of attribute '{name}' in '{self._spec.name}' "
                    f"sub-analysis as it is mapped to '{mapped_name}' in the parent "
                    f"analysis {self._parent}"
                )
            setattr(self._analysis, name, value)

    def __attrs_post_init__(self):
        """Ensure any nested subanalyses reference this subanalysis as the parent
        instead of the wrapped analysis object in order to link up chains of name
        mappings
        """
        for spec in self._analysis.__spec__.subanalysis_specs:
            subanalysis = getattr(self._analysis, spec.name)
            subanalysis._parent = self



from __future__ import annotations
import typing as ty
import attrs
from arcana.core import PACKAGE_NAME, __version__, CODE_URL
from arcana.core.utils.misc import fromdict_converter


class Metadata:
    pass


@attrs.define
class ContainerMetadata(Metadata):

    type: ty.Optional[str] = None
    tag: ty.Optional[str] = None
    uri: ty.Optional[str] = None


@attrs.define
class GeneratorMetadata(Metadata):

    name: str = PACKAGE_NAME
    version: str = __version__
    description: str = f"Empty dataset created by {PACKAGE_NAME}"
    code_url: str = CODE_URL
    container: ContainerMetadata = attrs.field(
        default=None, converter=fromdict_converter(ContainerMetadata)
    )


@attrs.define
class SourceDatasetMetadata(Metadata):

    url: ty.Optional[str] = None
    doi: ty.Optional[str] = None
    version: ty.Optional[str] = None


DEFAULT_README = f"""
This dataset was specified/generated using the Arcana analysis workflows framework
{CODE_URL}. However, no specific README was provided so this template is used instead.

The dataset could have been defined on already existing data, using

    $ arcana dataset ...

or via the Python API
"""


@attrs.define(kw_only=True)
class DatasetMetadata(Metadata):
    """Metadata describing the dataset. Based on the Brain Imaging Data Structure (BIDS)
    specification for dataset metadata

    Parameters
    ----------
    name : str
        human interpretable name/title of the dataset
    generated_by : list[GeneratorMetadata]
        list of tools used to generate the dataset in its current form
    acknowledgements : str
        Acknowledgements of non-author contributions to the dataset
    authors : list[str]
        list of authors who contributed to the preparation of the dataset
    doi : str, optional
        the DOI for the dataset
    funding : list[str]
        acknowledgements of funding bodies) associated with the dataset
    license : str
        usage license for the dataset
    references : list[str]
        publications associated with the dataset
    how_to_acknowledge : str
        details on how to acknowledge the dataset in publications
    ethics_approvals : list[str]
        list of ethic approvals associated with the dataset
    sources : list[SourceDatasetMetadata]
        for derivative datasets, a list of source dataset used to generate the derivative
    description : str
        a description of the dataset
    type : str
        type of dataset, by default "derivative"
    row_metadata : dict[str, ty.Dict[str, str]]
        a dictionary of dictionaries that map keys of metadata present in the rows of the
        dataset with a short description. The key of the outer dictionary is the frequency
        of the rows
    """

    name: str = attrs.field(default="Autogenerated-dataset")
    generated_by: ty.List[GeneratorMetadata] = attrs.field(
        converter=fromdict_converter(ty.List[GeneratorMetadata])
    )
    acknowledgements: str = attrs.field(default="Generic BIDS dataset", repr=False)
    authors: ty.List[str] = attrs.field(factory=list, repr=False)
    doi: str = attrs.field(default=None, repr=False)
    funding: ty.List[str] = attrs.field(factory=list, repr=False)
    license: str = attrs.field(default="CC0", repr=False)
    references: ty.List[str] = attrs.field(factory=list)
    how_to_acknowledge: str = attrs.field(default="see licence")
    ethics_approvals: ty.List[str] = attrs.field(factory=list)
    sources: ty.List[SourceDatasetMetadata] = attrs.field(
        factory=list, converter=fromdict_converter(ty.List[SourceDatasetMetadata])
    )
    description: str = attrs.field(default=DEFAULT_README)
    type: str = attrs.field(default="derivative", repr=False)
    row_metadata: ty.Dict[str, ty.Dict[str, str]] = attrs.field(factory=dict)

    @generated_by.default
    def generated_by_default(self):
        return [GeneratorMetadata()]

    # def tobids(self):
    #     dct = {}
    #     dct["Name"] = self.name
    #     dct["Acknowledgements"] = self.acknowledgements
    #     dct["Authors"] = self.authors
    #     if self.doi:
    #         dct["DOI"] = self.doi
    #     dct["Funding"] = self.funding
    #     dct["License"] = self.license
    #     dct["References"] = self.references
    #     dct["HowToAcknowledge"] = self.how_to_acknowledge
    #     dct["EthicsApprovals"] = self.ethics_approvals
    #     dct["Readme"] = self.description
    #     dct["GeneratedBy"] = [gb.tobids() for gb in self.generated_by]
    #     dct["Sources"] = [s.tobids() for s in self.sources]
    #     return dct

    # @classmethod
    # def frombids(cls, dct):
    #     return cls(
    #         name=dct.get("Name"),
    #         acknowledgements=dct.get("Acknowledgements"),
    #         authors=dct.get("Authors"),
    #         doi=dct.get("DOI"),
    #         funding=dct.get("Funding"),
    #         license=dct.get("License"),
    #         references=dct.get("References"),
    #         how_to_acknowledge=dct.get("HowToAcknowledge"),
    #         ethics_approvals=dct.get("EthicsApprovals"),
    #         description=dct.get("Readme"),
    #         generated_by=[GeneratorMetadata(gb) for gb in dct["GeneratedBy"]],
    #         sources=[SourceDatasetMetadata(s) for s in dct["Sources"]],
    #     )


def metadata_converter(metadata):
    if not metadata:
        metadata = {}
    elif not isinstance(metadata, DatasetMetadata):
        metadata = DatasetMetadata(**metadata)
    return metadata
